#!/usr/bin/perl
#
#  Wedge Martin
#
#
#  Licensed Materials - Property of IBM
#  5724-P69
#  <A9> Copyright IBM Corporation 2005, 2009. All Rights Reserved.
#  US Government Users Restricted Rights- Use, duplication or disclosure
#  restricted by GSA ADP Schedule Contract with IBM Corp.
#
#  Process binary AMA files and XML IPDR files to create CSV for
#  the UBA to process.
#
#  Replaces Perl sort with a linked list connecting all of the AMA records in the
#  recordHash
#
#  $Id: cs2kcdr_ng.pl 5207 2009-5-12 08:14:00Z wmartin $
#               1.0 2009-5-18 20:00 Jon Slater - Amended to reflect BT formula npaths
#               1.1 2009-5-20 19:00 Jon Slater - Amended to provide Lost Packet Percentage
#
#
# PDR format looks like     QCA.{nni}.QoS.{yyyy}.{mm}.{dd}_{hh}.{mm}_GMT.xml.gz QCA.9176524.QoS.2007.08.23_15.36_GMT.xml.gz
# AMA format looks like   {nni}_U{yy}{mm}{dd}{hh}{mm}{ss}AMA.unp  9176516_U070823024935AMA.unp
# keys
# nni  = the sensor ID for the switch - This is the same number that is contained in the AMA record.
# yyyy = 2007
# yy = 07
# mm = 08
# dd = 23
# hh = 13
# mm = 02
# ss = seconds
# pro = processed

use strict;

use POSIX qw(setsid);


$ENV{'TZ'} = "GMT";

#use Net::FTP;
use Date::Manip;
use Data::Dumper;
#use /usr/perl5/site_perl/5.6.1/sun4-solaris-64int/Date/Date::Manip;
use File::Copy;
use File::Basename;
use constant TRACK_EXTENSIONS => 1;


### AMA variables


my $totalRecords = 0;
my $totalIpdr = 0;
my $logType;
my $recordHash;
my $currentRecord;
my $startEpoch;
my $launchpath = $0;
my $lastAMAstamp;
my $aboveMOScounter = 0;
my $qosCallCounter = 0;
my $ama_array;

my $location_cache = {};

my $start_record = { timestamp => -1 };
my $stop_record = { timestamp => 99999999999999 };

$start_record->{next_record} = $stop_record;
$stop_record->{previous_record} = $start_record;


$SIG{'HUP'} = sub {
   open(TMP_HUP, "> /tmp/cs2k.dump");
   print TMP_HUP Dumper($recordHash);
   close(TMP_HUP);
};


##################################
###### USER CONFIG SECTION #######
##################################

# This has been moved to cs2kcdr-term.pl

BEGIN {
   my $pathname = dirname($0);
   require "$pathname/cs2kcdr-config.pl";
};

use vars qw(
   $logDir $NNI $maxFilesPerRun
   $runInterval $globalAggrInterval
   $numberOfOutputs $outputDir $exceptDir
   $storeExceptionDetail $DEBUG $MOSsla
   $terminationCodeHash $cdrConfig $trunkConfig $locationMap
   $amapath $amadir
   $ipdrpath $ipdrdir

);


my $structureCodeHash = {
     500 => {
           answer => 7,
           connect_time => 14,
           elapsed_time => 15,
           domestic_intl_indicator => 11,
           originating_open_digits1 => 9,
           originating_open_digits2 => 10,
           terminating_open_digits1 => 12,
           terminating_open_digits2 => 13,
           originating_charge_info => -1,
     },
     511 => {
           answer => 9,
           num_digits => 13,
           connect_time => 21,
           elapsed_time => 22,
           domestic_intl_indicator => 17,
           originating_open_digits1 => 14,
           originating_open_digits2 => 15,
           terminating_open_digits1 => 19,
           terminating_open_digits2 => 20,
           originating_charge_info => 16,
     },
     512 => {
           answer => 9,
           num_digits => 13,
           connect_time => 21,
           elapsed_time => 22,
           domestic_intl_indicator => 17,
           originating_open_digits1 => 14,
           originating_open_digits2 => 15,
           terminating_open_digits1 => 19,
           terminating_open_digits2 => 20,
           originating_charge_info => 16,
     },
     625 => {
           answer => 9,
           connect_time => 18,
           elapsed_time => 19,
           domestic_intl_indicator => 15,
           originating_open_digits1 => 13,
           originating_open_digits2 => 14,
           terminating_open_digits1 => 16,
           terminating_open_digits2 => 17,
           originating_charge_info => -1,
       },
       1 => {
           answer => 9,
           connect_time => 18,
           elapsed_time => 19,
           domestic_intl_indicator => 15,
           originating_open_digits1 => 13,
           originating_open_digits2 => 14,
           terminating_open_digits1 => 16,
           terminating_open_digits2 => 17,
           originating_charge_info => -1,
       }
};



##### END EDITABLE SECTION ####################
##### END EDITABLE SECTION ####################
##### END EDITABLE SECTION ####################

sub logDebug( $ ) {
   my $msg = shift;
   if ( $DEBUG ) {
       print STDERR "Debug: $msg\n";
   }
}


# my $logTimestamp = `date +%Y%m%d-%H%M%S`;
# chomp($logTimestamp);
my $logFile = "cs2kcdr_$NNI.log";

close(STDIN);
close(STDOUT);
close(STDERR);
open(STDIN, '/dev/null');
open(STDOUT, "> $logDir/$logFile");
open(STDERR, "> $logDir/error-$logFile");
print STDERR "Beginning cs2kcdr error log: " . time() . "\n";



########################################
##### function _add_to_list ######
########################################

sub _add_to_list( $ ) {
   my $rec = shift;
   if ( $rec->{next_record} || $rec->{previous_record} ) {
       # print "Warning:  Found duplicate callID for : " . $rec->{uci} . "\n";
       return;
   }
   if ( $rec->{timestamp} ) {
      my $test = $stop_record;
      # print "Adding: " . $rec->{timestamp} . "\n";
      while ( $test = $test->{previous_record} ) {
          if ( $rec->{timestamp} >= $test->{timestamp} ) {
                # print "Adding record...\n";
                my $next = $test->{next_record};
                $rec->{next_record} = $next;
                $rec->{previous_record} = $test;
                $test->{next_record} = $rec;
                $next->{previous_record} = $rec;
                last;
          }
      }
   } else {
      print "Warning: Unable to add AMA record to the list.  No timestamp found.\n";
      print Dumper($rec) . "\n";
   }
}


########################################
##### function _remove_from_list ######
########################################

sub _remove_from_list( $ ) {
   my $rec = shift;
   # print "Removing record from list: " . $rec . " = " . $rec->{timestamp} . "\n";
   my $pre = $rec->{previous_record};
   my $post = $rec->{next_record};
   $pre->{next_record} = $post;
   $post->{previous_record} = $pre;
   # cleanup
   $rec = undef;
}


########################################
##### function _getSubelementName ######
########################################

sub _getSubelementName( $ ) {
     my $type = shift;
     my $se_name = "CDR_";
     if ( $type eq "Zone" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_Zone_" . _getZone();
     } elsif ( $type eq "Path" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" .  "_Path_" . _getZone() . "_" . $currentRecord->{trunk_facility};
     } elsif ( $type eq "Device" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_Device";
     } elsif ( $type eq "TC" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_TC_";
          $se_name .= $terminationCodeHash->{$currentRecord->{facility_release_cause}} || $currentRecord->{facility_release_cause};
            print "[TC]  Termination Cause SE name is: $se_name\n";
     } elsif ( $type eq "TypeOfCall" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_TypeOfCall_" . $currentRecord->{typeofcall};
     } elsif ( $type eq "Extension1" ) {
          $se_name .= "Extension" . "_$NNI" . "_" . _mapExtension($currentRecord->{originating_open_digit1}, $currentRecord->{originating_open_digit2});
     } elsif ( $type eq "Extension2" ) {
          $se_name .= "Extension" . "_$NNI" . "_" . _mapExtension($currentRecord->{terminating_open_digit1}, $currentRecord->{terminating_open_digit2});
     } elsif ( $type eq "Trunk" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_Trunk_" . _getTrunkId();
     } elsif ( $type eq "Customer" ) {
          my $profile;
          $profile = "External" if ( $currentRecord->{originating_charge_info} eq "001" );
          $profile = "Residential" if ( $currentRecord->{originating_charge_info} eq "010" );
          $profile = "Corporate" if ( $currentRecord->{originating_charge_info} eq "020" );
          $profile = "InterTrunk" if ( $currentRecord->{originating_charge_info} eq "255" );
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_Customer_" . $profile;
     }
     return $se_name;
}



#####################################
##### function _getElementName ######
#####################################

sub _getElementName() {
    my $e_name;
    if ( $cdrConfig->{$currentRecord->{sensor_id}} ) {
        $e_name = $cdrConfig->{$currentRecord->{sensor_id}}->{name};
    } elsif ( $currentRecord->{sensor_id} ) {
        $e_name = $currentRecord->{sensor_id};
    } else {
        $e_name = "SoftswitchUnknown";
    }
    return $e_name;
}



#####################################
##### function _getCS2Kid ######
#####################################

sub _getCS2Kid( $ ) {
    my $id = shift;
    if ( $cdrConfig->{$id} ) {
         return $cdrConfig->{$id}->{name};
    } else {
         return $id;
    }
}


#####################################
##### function _getTrunkId ######
#####################################

sub _getTrunkId() {
    my $trunkId = substr($currentRecord->{trunk_facility}, 1, 4);
    if ( $trunkConfig->{$trunkId} ) {
          return $trunkConfig->{$trunkId}->{name};
    } else {
          return $trunkId;
    }
}



sub _amastamp( $ ) {
  my $fname = shift;
  my $file;
  ( $file ) = $fname =~ /(?:P|U)(.*)A/;
  my $year = substr( $file, 0, 2 );
  $year += 2000;
  my $month = substr( $file, 2, 2 );
  my $day = substr( $file, 4, 2 );
  my $hour = substr( $file, 6, 2 );
  my $minute = substr( $file, 8, 2 );
  my $second = substr( $file, 10, 2 );
  my $stamp;
  if ( $month ) {
     $stamp = Date_SecsSince1970( $month, $day, $year, $hour, $minute, $second );
  }
  return $stamp;
}


sub _amasort() {
   # AMA format looks like                 {nni}_P{yy}{mm}{dd}{hh}{mm}{ss}AMA.prO  9176516_P070823024935AMA.pro
    my $sorthash;
    my ( $a_file, $b_file );
    if ( $a && $b ) {
       $sorthash->{$a}->{timestamp} = _amastamp($a);
       $sorthash->{$b}->{timestamp} = _amastamp($b);
    }
    $sorthash->{$a}->{timestamp} <=> $sorthash->{$b}->{timestamp};
}


sub _ipdrstamp( $ ) {
   my $fname = shift;
   my $line = [ split(/\./, $fname ) ];
   my $year = $line->[3];
   my $month = $line->[4];
   my ( $day, $hour ) = split(/_/, $line->[5]);
   my ( $minute, undef ) = split(/_/, $line->[6]);
   my $stamp;
   if ( $month ) {
      $stamp = Date_SecsSince1970( $month, $day, $year, $hour, $minute, 0 );
   }
   return $stamp;
}

sub _ipdrsort() {
   my $sorthash;
   if ( $a && $b ) {
      # IPDR format looks like     QCA.{nni}.QoS.{yyyy}.{mm}.{dd}_{hh}.{mm}_GMT.xml.gz QCA.9176524.QoS.2007.08.23_15.36_GMT.xml.gz
      $sorthash->{$a}->{timestamp} = _ipdrstamp($a);
      $sorthash->{$b}->{timestamp} = _ipdrstamp($b);
   }

   $sorthash->{$a}->{timestamp} <=> $sorthash->{$b}->{timestamp};


}



sub _genAmaRecordTimestamp( $ $ ) {
   my $date = shift;
   my $time = shift;
   my $year;
   my $month;
   my $day;
   my $hour;
   my $minute;
   my $ts;
   my @dateArray = split(//, $date);
   $year = "200" . $dateArray[0];
   $month = $dateArray[1] . $dateArray[2];
   $day = $dateArray[3] . $dateArray[4];

   my @timeArray = split(//, $time);
   $hour = $timeArray[0] . $timeArray[1];
   $minute = $timeArray[2] . $timeArray[3];
   $ts = Date_SecsSince1970($month,$day,$year,$hour,$minute,0);
   return $ts;
}


#####################################
##### function _processAMA ######
#####################################

sub _processAMA() {

   my $skip_counter = 0;  # how many times we skipped calculating the date
   my $ama_date_counter = 0;
   my $tmp_timestamp_hash;
   $logType = "ama";
   my $newFileArr;

   opendir(AMA_STORE, $amadir) || die "Could not open dir '$amadir': $!\n";
   my @old_files = readdir(AMA_STORE);
   close(AMA_STORE);

   opendir(NEW_FILES, $amapath) || die "Could not open dir '$amapath': $!\n";
   my @predir = readdir(NEW_FILES);

   my $fileArray;
   my $fcounter = 0;
   # map { $_ = (split(/\s+/, $_)) } @predir;
   foreach my $listFile ( sort _amasort grep {$_ ne '.' && $_ ne '..' } @predir ) {
      push(@$fileArray, $listFile) if $listFile;
   }

   foreach my $f ( @$fileArray ) {
       if ( ! grep { $_ eq $f } @old_files ) {
            # we have a new file...
            if ( $fcounter < 1 ) {
                if ( $lastAMAstamp && _amastamp($f) < ( $lastAMAstamp + 900 ) ) {
                  $lastAMAstamp = _amastamp($f);
                  print "Downloading file: $f ... ";
                  # move("$amapath/$f", "$amadir/$f");
                  move("$amapath/$f", "$amadir/$f");
                  print "Done.\n";
                  push ( @$newFileArr, $f );
                } else {
                  if ( ! $lastAMAstamp ) {
                     $lastAMAstamp = _amastamp($f);
                  } else {
                     $lastAMAstamp += 900;
                  }
                  return 0;
                }
                $fcounter++;
            }
       } else {
            print "Skipping already obtained file: $f\n";
       }

   }


   my $timestampStart = time();
   if ( $newFileArr ) {
      print "Processing AMA files.... TIMESTAMP: $timestampStart\n";
   }
   foreach my $file ( @$newFileArr ) {
      my ( $data, $foo );
      open (AMAFILE, "< $amadir/$file") || die "Unable to open file '$file': $!\n";
      binmode(AMAFILE);
      while ( read(AMAFILE, $foo, 180000) ) {
          my $n1 = length($foo);
          my $s1 = 2 * $n1;
          my $datahere = unpack("H$s1", $foo);
          $datahere =~ tr/[a-f]/[A-F]/;
          $data .= $datahere;
      }

      my $recordArr = [ split(/(?:AA|AB)+/, $data) ];
      my ( $timestamp, $dataArr );
      my $callId;
      foreach my $record ( @$recordArr ) {
         $callId = "";
         $dataArr = [ split(/(?:FFFFFFFFFFFFFFFF|FFFFFFFFFF|FFFF|C)/, $record) ];
         next if ( scalar(@$dataArr) < 5 );
         next if ( ! grep { $_ eq "612" } @$dataArr );
         my $structure_code = $dataArr->[0];
         $structure_code = substr($structure_code, 2, 5);
         my $trunk_facility;
         my $facility_release_cause;
         if ( $dataArr->[0] =~ /^(4|5)/ ) {  #  only look for module codes if we get a '4' or '5' prefix to the structure code
            foreach ( my $marker = 22; $marker < 60; $marker++ ) {
                my $field = $dataArr->[$marker];
                if ( $field eq "612" && $dataArr->[$marker + 1] eq "8006600" ) {
                      if ( $callId ) {
                         #  we already have our first 612 module code for the record.  If we see additional ones
                         #    we check for matching IPDR and in the event that there is no match for the first
                         #     612, we map the additional 612 call Id's IPDR to the first 612 'index' id
                         my $addCall = _generateCallId($dataArr->[$marker +2] . $dataArr->[$marker +3]);
                         my $callKey = _generateCallId($callId);
                         if ( $recordHash->{$addCall}->{ipdr} && ! $recordHash->{$callKey}->{ipdr} ) {
                              _mapIpdrToCall( $callKey, $recordHash->{$addCall} );
                              if ( $recordHash->{$callKey}->{ipdr} ) {
                                  print "Mapped: $addCall ipdr data to $callKey\n";
                              }
                         }

                      }
                      $callId ||= $dataArr->[$marker + 2] . $dataArr->[$marker + 3];
                      $marker += 3;
                } elsif ( $field eq "104" ) {
                      $trunk_facility = $dataArr->[$marker + 1];
                      $marker++;
                } elsif ( $field eq "130" ) {
                #       print "[fac] " . join(', ', @$dataArr) . "\n";
                      $facility_release_cause = substr($dataArr->[$marker + 1], 2, 5);
                #       print "[facility_release_cause] $facility_release_cause\n";
                      $marker++;
                }
            }
         }
         $callId = _generateCallId( $callId );

         next if ( length($callId) < 20 );

         if ( $recordHash->{$callId} && $recordHash->{$callId}->{ama} == 1 ) {
            if ( $recordHash->{$callId}->{timing_indicator} =~ /000/ || ! $recordHash->{$callId}->{timing_indicator} ) {
               # 3 way call, etc.
               # if the timing indicator is 00100, 00200, or 00300, we have a 'long call' and will handle it by summing the call duration and flushing
               # the call out when we see the 00300 indicator, signifiying that the call is complete.
               my $ipdr;
               if ( $recordHash->{$callId}->{ipdr} == 1 ) {
                  $ipdr = $recordHash->{$callId};
               }
               $callId .= "_" . $dataArr->[$structureCodeHash->{$structure_code}->{connect_time}];
               if ( $ipdr ) {
                  _mapIpdrToCall( $callId, $ipdr );
               }
               next;
            }
         }


         $recordHash->{$callId}->{ama_file} = $file;
         $recordHash->{$callId}->{ama} = 1;
         $recordHash->{$callId}->{call_code} = $dataArr->[1];
         $recordHash->{$callId}->{sensor_id} = $dataArr->[3];
         $recordHash->{$callId}->{timing_indicator} = $dataArr->[7];
         $recordHash->{$callId}->{answer} = $dataArr->[$structureCodeHash->{$structure_code}->{answer}];
         $recordHash->{$callId}->{originating_open_digit1} = $dataArr->[$structureCodeHash->{$structure_code}->{originating_open_digits1}];
         $recordHash->{$callId}->{originating_open_digit2} = $dataArr->[$structureCodeHash->{$structure_code}->{originating_open_digits2}];
         $recordHash->{$callId}->{originating_charge_info} = $dataArr->[$structureCodeHash->{$structure_code}->{originating_charge_info}];
         $recordHash->{$callId}->{domestic_intl_indicator} = $dataArr->[$structureCodeHash->{$structure_code}->{domestic_intl_indicator}];
         $recordHash->{$callId}->{terminating_open_digit1} = $dataArr->[$structureCodeHash->{$structure_code}->{terminating_open_digits1}];
         $recordHash->{$callId}->{terminating_open_digit2} = $dataArr->[$structureCodeHash->{$structure_code}->{terminating_open_digits2}];
         $recordHash->{$callId}->{connect_time} = $dataArr->[$structureCodeHash->{$structure_code}->{connect_time}];
         my $elapsed_time = $dataArr->[$structureCodeHash->{$structure_code}->{elapsed_time}];
         if ( $elapsed_time > 0 ) {
            my ( $minutes, $seconds ) = $elapsed_time =~ /0(\d{5})(\d{3})/;
             # seconds are actually 10ths here. we need to divide them by 10
            my $realseconds = ( 60 * $minutes ) + ( $seconds / 10 );
            $recordHash->{$callId}->{elapsed_time} += $realseconds;
         }
         $recordHash->{$callId}->{uci} = $callId;
         if ( !  $recordHash->{$callId}->{answer}  || $recordHash->{$callId}->{answer} == 0 ) {
            $recordHash->{$callId}->{answer} = 1;
         } else {
            $recordHash->{$callId}->{answer} = 2;
         }
         $recordHash->{$callId}->{cs2kID} = _getCS2Kid($recordHash->{$callId}->{sensor_id});
         $recordHash->{$callId}->{trunk_facility} = $trunk_facility;
         $recordHash->{$callId}->{facility_release_cause} = $facility_release_cause;
         my $trunk_type = substr($trunk_facility, 0, 1);
         my $ctype = $recordHash->{$callId}->{call_code};
       #  if ( $ctype eq "901" ) {
       #      $ctype = "Internal";
       #  } elsif ( $trunk_type == 1 ) {
       #      $ctype = "Incoming";
       #  } elsif ( $trunk_type == 2 ) {
       #      $ctype = "Outgoing";
       #  }
         $recordHash->{$callId}->{typeofcall} = $ctype;
         if ( ! $tmp_timestamp_hash->{$dataArr->[6] . $dataArr->[$structureCodeHash->{$structure_code}->{connect_time}]} ) {
            $ama_date_counter++;
            $timestamp = _genAmaRecordTimestamp( $dataArr->[6], $dataArr->[$structureCodeHash->{$structure_code}->{connect_time}] );
            $tmp_timestamp_hash->{$dataArr->[6] . $dataArr->[$structureCodeHash->{$structure_code}->{connect_time}]} = $timestamp;
         } else {
            $skip_counter++;
            $timestamp = $tmp_timestamp_hash->{$dataArr->[6] . $dataArr->[$structureCodeHash->{$structure_code}->{connect_time}]};
         }
         $recordHash->{$callId}->{timestamp} = $timestamp + $recordHash->{$callId}->{elapsed_time};
            my $rec_term_day = &UnixDate("epoch " . $recordHash->{$callId}->{timestamp}, "%d");
            $recordHash->{$callId}->{rec_term_day} = $rec_term_day;
         if ( $recordHash->{$callId}->{timestamp} =~ /\./ ) {
             $recordHash->{$callId}->{timestamp} = (split(/\./, $recordHash->{$callId}->{timestamp}))[0];
         }
         if ( ! $recordHash->{$callId}->{elapsed_time} ) {
              $recordHash->{$callId}->{elapsed_time} = "0.0";
         }
         if ( $recordHash->{$callId}->{timing_indicator} !~ /00100/ && $recordHash->{$callId}->{timing_indicator} !~ /00200/ ) {
            #  we make sure it's not a 100 or 200 as those are start stamps and increments for long duration calls, respectively.
            #  for long duration calls, we only want to put it into the array for the last record, the 00300.
            if ( ! $recordHash->{$callId}->{ipdr} ) {
                    $recordHash->{$callId}->{RAW} = join(', ', @$dataArr);
            }
            $totalRecords++;
            push(@$ama_array, $recordHash->{$callId});
          }
       }
       close(AMAFILE);
   }
   print "Finished processing AMA data... " . time() . "\n";
   print " called Date::Manip $ama_date_counter times and skipped it $skip_counter times.\n";
   return 1;
}



sub _currentCallIsComplete() {
    # print "DUMPER_in_currentCallIsComplete(): " . Dumper($currentRecord, undef) . "\n\n";
    if ( $currentRecord->{ama} == 1 && $currentRecord->{ipdr} == 1 ) {
        return 1;
    }
}


sub _mapIpdrToCall( $ $ ) {
     my $call = shift;
     my $ipdr = shift;
     my $metrics = [ 'averagePacketLatency', 'inboundPacketCount', 'inboundByteCount', 'outboundByteCount', 'inboundLostPacketCount',
                      'packetDelayVariation', 'AverageLossRate', 'ResidualEchoReturnLoss', 'EndSystemDelay', 'MOSLQ', 'RoundTripTimeHighWaterMark' ];
     $recordHash->{$call}->{ipdr} = 1;
     $recordHash->{$call}->{uci} = $call;
     for my $m ( @$metrics ) {
        $recordHash->{$call}->{$m} = $ipdr->{$m};
     }
}


sub _getZone() {
   if ( $currentRecord->{sensor_id} ) {
       if ( $cdrConfig->{$currentRecord->{sensor_id}} ) {
          return $cdrConfig->{$currentRecord->{sensor_id}}->{zone};
       } else {
          return $currentRecord->{sensor_id};
      }
   } elsif ( $currentRecord->{subscriberId} ) {
      if ( $currentRecord->{subscriberId} =~ /aaln/ ) {
           if ( $currentRecord->{subscriberId} =~ /\./ ) {
              my $subscriber = (split(/\./, $currentRecord->{subscriberId}))[0];
              return $subscriber;
          } else {
              return $currentRecord->{subscriberId};
          }
      }
   }
}


sub _getSoftswitchLocation( $ ) {
   if ( $cdrConfig->{$currentRecord->{sensor_id}} ) {
        return $cdrConfig->{$currentRecord->{sensor_id}}->{location};
   }
   return "Unmapped";
}


sub _getLocation( $ $ ) {
   my ( $data, $type ) = @_;
   my $stripDigits;
   return $data if ( $data eq "Unknown" );
   if ( $location_cache->{$data} ) {
      return $location_cache->{$data};
   }
   foreach my $loc ( keys %$locationMap ) {
       if ( $type eq "extension" ) {
           my $regexp = $locationMap->{$loc}->{extensionPattern};
           if ( $regexp && $data =~ $locationMap->{$loc}->{extensionPattern} ) {
                ( $stripDigits ) = $data =~ /$regexp(.*)/;
                my @stripDigitArr = split(//, $stripDigits);
                for ( my $r = 0; $r < scalar(@{$locationMap->{$loc}->{extensionRange}}); $r++ ) {
                     my $range = $locationMap->{$loc}->{extensionRange}->[$r];
                     my $sizedLastDigits;
                     if ( $range =~ /-/ ) {
                         my @nums = split(/-/, $range);
                         for ( my $i = 0; $i < length($nums[0]); $i++ ) {
                             $sizedLastDigits .= $stripDigitArr[$i];
                         }
                         if ( $sizedLastDigits >= $nums[0] && $sizedLastDigits <= $nums[1] ) {
                               $location_cache->{$data} = $loc;
                               return $loc;
                         }
                     } else {
                         for ( my $i = 0; $i < length($range); $i++ ) {
                               $sizedLastDigits .= $stripDigitArr[$i];
                         }
                         if ( $sizedLastDigits eq $range ) {
                               $location_cache->{$data} = $loc;
                               return $loc;
                         }
                     }
                }
           }
      } elsif ( $type eq "ipAddress" ) {
           my $returnLoc = "false";
           my $ip = $data;
           if ( $locationMap->{$loc}->{ipAddress} ) {
               my @maskArray = split(/\./, $locationMap->{$loc}->{ipAddress});
               my @ipArray = split(/\./, $ip);
               for ( my $p = 0; $p < scalar(@maskArray); $p++ ) {
                   if ( $maskArray[$p] =~ /-/ ) {
                       my @stoppers = split(/-/, $maskArray[$p]);
                       if ( $ipArray[$p] >= $stoppers[0] && $ipArray[$p] <= $stoppers[1] ) {
                            $returnLoc = "true";
                       }
                   } else {
                       if ( $ipArray[$p] == $maskArray[$p] ) {
                            $returnLoc = "true";
                       } else {
                            $returnLoc = "false";
                       }
                   }
               }
           }
           if ( $returnLoc eq "true" ) {
               $location_cache->{$data} = $loc;
               return $loc;
           }
      }
   }
   $location_cache->{$data} = "Unmapped";
   return "Unmapped";
}




sub _mapExtension( $ $ ) {
   my ( $field1, $field2 ) = @_;
   my $extension;
   if ( ! $field1 && ! $field2 ) {
        return "Untracked";
   }
   my $check;
   $extension = $field1;
   if ( $field1 !~ /^0/ && $field2 ) {
      $extension .= $field2;
   }
   my $locTest = _getLocation($extension, "extension");
   if ( $locTest !~ /Unknown/ && $locTest !~ /Unmapped/ ) {
       return $locTest;
   } else {
       return "Untracked";
   }
}



sub _isOffNetCall() {
    if ( $currentRecord->{trunk_facility} ) {
         return 1;
    } else {
         return 0;
    }
}

sub _isCallCode( $ ) {
   my $cc = shift;
   if ( $currentRecord->{call_code} == $cc ) {
       return 1;
   } else {
       return 0;
   }
}

sub _getElapsedTime() {
     if ( $currentRecord->{elapsed_time} ) {
         return $currentRecord->{elapsed_time};
     }
}


sub _isShortCall() {
    if ( 2 > $currentRecord->{elapsed_time} ) {
        return 1;
    }
}


sub _isSuccessfulCall() {
    if ( $currentRecord->{answer} == 1 ) {
        return 1;
    } else {
   return '0.0' ;
 }
}


sub _isUnsuccessfulCall() {
   if ( $currentRecord->{answer} == 2 ) {
        return 1;
   } else {
 return '0.0';
 }
}


sub _isFacilityReleaseCause( $ ) {
    my $frc = shift;
    if ( $currentRecord->{facility_release_cause} eq $frc ) {
         return 1;
    }
      if ( $frc eq "002" || $frc == 2 ) {
         return '0.0';
    }
}

sub _isCallCompletion( $ ) {
    my $cc = shift;
    if ( $currentRecord->{call_completion} == $cc ) {
        return 1;
    }
}


sub _getPacketLatency() {
   return $currentRecord->{averagePacketLatency};
}

sub _getInboundPackets() {
   return $currentRecord->{inboundPacketCount};
}

sub _getOutboundPackets() {
   return $currentRecord->{outboundPacketCount};
}

sub _getInboundBytes() {
   return $currentRecord->{inboundByteCount};
}

sub _getOutboundBytes() {
   return $currentRecord->{outboundByteCount};
}

sub _getLostPackets() {
   return $currentRecord->{inboundLostPacketCount};
}
#Added for BT by Jon Slater 20 May 2009
sub _getPacketLoss() {
        return $currentRecord->{packetLossPercentage};
}

sub _getPacketDelayVariation() {
# return $currentRecord->{packetDelayVariation};
 if ( $currentRecord->{packetDelayVariation} ) {
      return $currentRecord->{packetDelayVariation};
   } else {
       return '0.0';
 }
}


sub _getAboveMOSpercentage() {
   # print STDERR "getAboveMOSpercentage: 100 * ( $aboveMOScounter / $qosCallCounter )";
   if ( $aboveMOScounter == 0 ) {
      return 0;
   }
   return 100 * ( $aboveMOScounter / $qosCallCounter );
}

sub _getAverageLossRate() {
   return $currentRecord->{AverageLossRate};
}

sub _getEchoReturnLoss() {
   return $currentRecord->{ResidualEchoReturnLoss};
}

sub _getEndSystemDelay() {
   return $currentRecord->{EndSystemDelay};
}


sub _getListeningQualityMOS() {
   return $currentRecord->{MOSLQ};
}

sub _getRoundTripAverage() {
   return $currentRecord->{RoundTripTimeHighWaterMark};
}

sub _getTransmissionStreams() {
   return $qosCallCounter;
}

sub _getUnderMOSLimit() {
   if ( $currentRecord->{MOSLQ} < $MOSsla && $currentRecord->{MOSLQ} ) {
      return 1;
   }
}

sub _getUnderMOSLimitOffnet() {
   if ( $currentRecord->{MOSLQ} < $MOSsla && _getLocation($currentRecord->{LocalMediaIpAddress}, "extension") =~ /Unmapped/ && $currentRecord->{MOSLQ} ) {
      return 1;
   }
}

sub _getUnderMOSLimitOnnet() {
   if ( $currentRecord->{MOSLQ} < $MOSsla && _getLocation($currentRecord->{LocalMediaIpAddress}, "extension") !~ /Unmapped/ && $currentRecord->{MOSLQ} ) {
      return 1;
   }
}






my $functionMap = {

    "TotalCall" =>                      { aggr => "sum", type =>  "ama",  method =>  sub { return 1 if $currentRecord->{ama} == 1 } },
    # "IncomingCall890" =>                { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(890) } },
    # "IncomingCall891" =>                { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(891) } },
    "IncomingCall" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return (  _isCallCode(891) || _isCallCode(890) ) } },
    "OutgoingCall" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(6) } },
    "OnNetCall" =>                      { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(901) } },
    "OffNetCall" =>                     { aggr => "sum", type =>  "ama",  method =>  sub { return _isOffNetCall() } },
    "InternalCall" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(901) } },
    "MobileCall" =>                     { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(801) } },
    "InternationalCall" =>              { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(800) } },
    "NationalCall" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(805) } },
    "SuccessfulCall" =>                 { aggr => "sum", type =>  "ama",  method =>  sub { return _isSuccessfulCall() } },
    "UnsuccessfulCall" =>               { aggr => "sum", type =>  "ama",  method =>  sub { return _isUnsuccessfulCall() } },
    "UnsuccessfulCallInvalidAddress" => { aggr => "sum", type =>  "ama",  method =>  sub { return _isFacilityReleaseCause("517") } },
    "UnsuccessfulCallNoAnswer" =>       { aggr => "sum", type =>  "ama",  method =>  sub { return _isFacilityReleaseCause("012") } },
    "UnsuccessfulCallBusy" =>           { aggr => "sum", type =>  "ama",  method =>  sub { return _isFacilityReleaseCause("001") } },
    "UnsuccessfulCallNetwork" =>        { aggr => "sum", type =>  "ama",  method =>  sub { return _isFacilityReleaseCause("002") } },
    "CallDuration" =>                   { aggr => "average", type =>  "ama",  method =>  sub { return _getElapsedTime() } },
    "NormalCallTermination" =>          { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCompletion("CC") } },
    "DropCall" =>                       { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCompletion("CAD") } },
    "ShortCall" =>                      { aggr => "sum", type =>  "ama",  method =>  sub { return _isShortCall() } },
    "PacketLatency" =>                  { aggr => "average", type =>  "ipdr", method =>  sub { return _getPacketLatency() } },
    "InboundPacket" =>                  { aggr => "sum", type =>  "ipdr", method =>  sub { return _getInboundPackets() } },
    "OutboundPacket" =>                 { aggr => "sum", type =>  "ipdr", method =>  sub { return _getOutboundPackets() } },
    "InboundByte" =>                    { aggr => "sum", type =>  "ipdr", method =>  sub { return _getInboundBytes() } },
    "OutboundByte" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _getOutboundBytes() } },
    "LostPacket" =>                     { aggr => "sum", type =>  "ipdr", method =>  sub { return _getLostPackets() } },
    "LostPacketpercentage" =>                   { aggr => "average", type => "ipdr", method => sub { return _getPacketLoss() } },
    "PacketDelayVariation" =>           { aggr => "average", type =>  "ipdr", method =>  sub { return _getPacketDelayVariation() } },
    # new extended qos in IPDR
    "AboveMOSpercentage" =>            { aggr => "average", type =>  "ipdr", method =>  sub { return _getAboveMOSpercentage() } },
    "AverageLossRate" =>               { aggr => "average", type =>  "ipdr", method =>  sub { return _getAverageLossRate() } },
    "EchoReturnLoss" =>                { aggr => "average", type =>  "ipdr", method =>  sub { return _getEchoReturnLoss() } },
    "EndSystemDelay" =>                { aggr => "average", type =>  "ipdr", method =>  sub { return _getEndSystemDelay() } },
    "ListeningQualityMOS" =>           { aggr => "average", type =>  "ipdr", method =>  sub { return _getListeningQualityMOS() } },
    "RoundTripAverage" =>              { aggr => "average", type =>  "ipdr", method =>  sub { return _getRoundTripAverage() } },
    "TransmissionStreams" =>           { aggr => "sum", type =>  "ipdr", method =>  sub { return _getTransmissionStreams() } },
    "UnderMOSLimit" =>         { aggr => "sum", type =>  "ipdr", method =>  sub { return _getUnderMOSLimit() } },
    "UnderMOSLimitOffnet" =>      { aggr => "sum", type =>  "ipdr", method =>  sub { return _getUnderMOSLimitOffnet() } },
    "UnderMOSLimitOnnet" =>       { aggr => "sum", type =>  "ipdr", method =>  sub { return _getUnderMOSLimitOnnet() } }

};



my $deviceMetrics = {
      Trunk =>       { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Media Gateway",
                    filter =>  sub { return 1 if ( length($currentRecord->{trunk_facility}) > 4 || _currentCallIsComplete() ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_MGW", VOIPSoftswitchVendor =>  "Nortel",
                                  VOIPSoftswitchZone =>  sub { return _getZone() },
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "IncomingCall890", "IncomingCall891", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork",
                                   "DropCall", "CallDuration", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                   "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                   "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet" ]
                  },
      TypeOfCall =>  { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Type of Call",
                    filter =>  sub { return 1 if (  ( $currentRecord->{call_code} > 0 && $currentRecord->{sensor_id} > 0 ) || _currentCallIsComplete() ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_TypeOfCall", VOIPSoftswitchVendor =>  "Nortel",
                                  VOIPSoftswitchZone =>  sub { return _getZone() },
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "MobileCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork",
                                   "DropCall", "CallDuration", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                   "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                   "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet" ]
                  },
       Customer =>    { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Customer Profile",
                     filter =>  sub { return 1 if ( ( $currentRecord->{call_code} > 0 && $currentRecord->{sensor_id} > 0 && $currentRecord->{originating_charge_info} > 0 ) || _currentCallIsComplete() ) },
                     properties =>  { VOIPSoftswitchComponentType =>  "CDR_Softswitch_CustomerProfile", VOIPSoftswitchVendor =>  "Nortel",
                                   VOIPSoftswitchZone =>  sub { return _getZone() },
                                   VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                     metricArray =>  [ "TotalCall", "IncomingCall", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                    "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                    "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork", "NormalCallTermination",
                                    "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket", "InboundByte", "OutboundByte",
                                    "LostPacket", "PacketDelayVariation", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                    "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                    "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet" ]
                   },
       Path =>        { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Path",
                     filter =>  sub { return 1 if ( _currentCallIsComplete() ) },
                     properties =>  { VOIPSoftswitchComponentType =>  "CDR_Path", VOIPSoftswitchVendor =>  "Nortel",
                                   VOIPSoftswitchZone =>  sub { return _getZone() },
                                   VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                     metricArray =>  [ "TotalCall", "MobileCall", "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall",
                                    "UnsuccessfulCallInvalidAddress", "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork",
                                    "NormalCallTermination", "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket",
                                    "InboundByte", "OutboundByte", "LostPacket", "PacketDelayVariation", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                    "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                    "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet" ]
                   },
       Zone =>        { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Zone",
                     filter =>  sub { return 1 if ( _currentCallIsComplete() ) },
                     properties =>  { VOIPSoftswitchComponentType =>  "CDR_Zone", VOIPSoftswitchVendor =>  "Nortel",
                                   VOIPSoftswitchZone =>  sub { return _getZone() },
                                   VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                     metricArray =>  [ "TotalCall", "MobileCall", "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall",
                                    "UnsuccessfulCallInvalidAddress", "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork",
                                    "NormalCallTermination", "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket",
                                    "InboundByte", "OutboundByte", "LostPacket", "PacketDelayVariation", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                    "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                    "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet" ]
                   },
      Device =>      { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Softswitch Device",
                    filter =>  sub { return 1 },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Softswitch_Device", VOIPSoftswitchVendor =>  "Nortel",
                                  VOIPSoftswitchZone =>  sub { return _getZone() },
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "IncomingCall890", "IncomingCall891", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork", "NormalCallTermination",
                                   "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket", "InboundByte", "OutboundByte",
                                   "LostPacket", "PacketDelayVariation", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                   "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                   "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet", "LostPacketpercentage" ]
                  },
      TC =>          { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Termination Cause",
                    filter =>  sub {return 1 if ( $currentRecord->{facility_release_cause} ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_TC", VOIPSoftswitchVendor =>  "Nortel",
                                  VOIPSoftswitchZone =>  sub { return _getZone() },
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "IncomingCall890", "IncomingCall891", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                   "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                   "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet" ]
                  },
     Extension1 =>   { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Extension",
                filter =>  sub { return 1 if ( TRACK_EXTENSIONS && ( $logType == "ama" || ( $logType == "ipdr" &&  _currentCallIsComplete() ) ) ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Extension", VOIPSoftswitchVendor =>  "Nortel",
                                  VOIPSoftswitchZone =>  sub { return _getZone() },
                                  VOIPSoftswitchLocation =>  sub { return _getLocation(_mapExtension($currentRecord->{originating_open_digit1}, $currentRecord->{originating_open_digit2} ), "extension" ) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "IncomingCall890", "IncomingCall891", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork", "NormalCallTermination",
                                   "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket", "InboundByte", "OutboundByte",
                                   "LostPacket", "PacketDelayVariation", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                   "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                   "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet", "LostPacketpercentage" ]
                  },
     Extension2 =>   { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Extension",
                filter =>  sub { return 1 if ( TRACK_EXTENSIONS && ( $logType == "ama" || ( $logType == "ipdr" &&  _currentCallIsComplete() ) ) ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Extension", VOIPSoftswitchVendor =>  "Nortel",
                                  VOIPSoftswitchZone =>  sub { return _getZone() },
                                  VOIPSoftswitchLocation =>  sub { return _getLocation(_mapExtension($currentRecord->{terminating_open_digit1}, $currentRecord->{terminating_open_digit2} ), "extension" ) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "IncomingCall890", "IncomingCall891", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork", "NormalCallTermination",
                                   "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket", "InboundByte", "OutboundByte",
                                   "LostPacket", "PacketDelayVariation", "ShortCall", "AboveMOSpercentage", "AverageLossRate", "EchoReturnLoss",
                                   "EndSystemDelay",  "ListeningQualityMOS", "RoundTripAverage", "TransmissionStreams",
                                   "UnderMOSLimit", "UnderMOSLimitOffnet", "UnderMOSLimitOnnet", "LostPacketpercentage" ]
                  }

};



sub _checkExceptionCallDetail() {
    foreach my $code ( @{$storeExceptionDetail} ) {
        if ( $currentRecord->{facility_release_cause} == $code ) {
           return &UnixDate("epoch " . $currentRecord->{timestamp}, "%Y%m%d-%H%M%S");
        }
    }
}



sub _flushData() {
    my $datafile;
    my $lagfile;
    my $last_timestamp;
    my $exception_time;

    # my $flush_skip_counter = 0;
    my $filecounter = 1;
    print "TotalRecords variable reports: $totalRecords\n";
    print ">>> Beginning data flush to file... " . time() . "\n";

    my $aggrHash;
    my $startAggrTime;
    my $exceptionFile;

    ## PERF
    my $aggrCounter = 0;

    # my $unixdate_counter = 0;

    my $deviceCalls = 0;
    my $call_code_hash = {};
    $currentRecord = undef;

    my $time_complete;
    my $last_record_day;
    # my $utime = &UnixDate("epoch $lastAMAstamp", "%Y_%m_%d_%H_%M_%S");
    # open(UNMATCHED, "> $logDir/cs2kcdr_unmatched_$utime.log");
    while ( $currentRecord = shift @$ama_array ) {

          if ( $currentRecord->{timestamp} < ( $last_timestamp - 3600 ) ) {
              print "WARNING: Found very old timestamp for record: " . $currentRecord->{uci} . "\n";
              print "    Found very old timestamp for a record: " . &UnixDate("epoch " . $currentRecord->{timestamp}, "%Y-%m-%d %H:%M:%S") . "\n";
              print "    is older than: " . &UnixDate("epoch " . ( $last_timestamp - 3600 ) , "%Y-%m-%d %H:%M:%S") . "\n";
          }
          if ( ! $startAggrTime ) {
             $startAggrTime = $lastAMAstamp;
             $time_complete = &UnixDate("epoch $startAggrTime", "%Y-%m-%d %H:%M:%S");
             print "Beginning aggregation period for file: $lastAMAstamp with time_complete: $time_complete\n";
          }


          if ( $storeExceptionDetail ) {
              my $exception_time = _checkExceptionCallDetail();
              if ( $exception_time ) {
                  if ( ! $exceptionFile ) {
                      #Amended exceptions output directory J Slater
                      open (EXCEPTIONS, "> $exceptDir/CDRexception_" . "$NNI" . "_log_$exception_time");
                      $exceptionFile = \*EXCEPTIONS;
                  }
                  $time_complete = &UnixDate("epoch " . $currentRecord->{timestamp}, "%Y-%m-%d %H:%M:%S");
                  print $exceptionFile $time_complete . "," . $currentRecord->{uci} . "," . $currentRecord->{facility_release_cause} . ","
                         . "," . $currentRecord->{call_code} . "," . $currentRecord->{elapsed_time} . "," . $currentRecord->{trunk_facility}
                         . "," . $currentRecord->{sensor_id} . "," . $currentRecord->{answer} .  "," . $currentRecord->{PacketSent} . ","
                         . $currentRecord->{OutboundPacket} . "," . $currentRecord->{LostPacket} . "," . $currentRecord->{PacketDelayVariation}
                         . "," . $currentRecord->{PacketLatency} . "\n";
              }
          }

          my $element = _getElementName();
          my $metric;
          my $aggr;
          my $subelement;
          my $val;
          my $agg_day;
          $agg_day = &UnixDate("epoch " . $startAggrTime, "%d");

          if ( ( $currentRecord->{timestamp} >= ( $startAggrTime - 900 ) ) && ( $currentRecord->{timestamp} < ( $startAggrTime + $globalAggrInterval ) ) ) {
                if($currentRecord->{rec_term_day} <= $agg_day) {
                    $currentRecord->{timestamp} = $startAggrTime;
                } elsif($currentRecord->{rec_term_day} > $agg_day) {
                    $startAggrTime = $startAggrTime+300;
                    $time_complete = &UnixDate("epoch " . $startAggrTime, "%Y-%m-%d %H:%M:%S");
                }

          } elsif ( ( $currentRecord->{timestamp} > ( $startAggrTime  + $globalAggrInterval ) ) || $currentRecord->{rec_term_day} ne $last_record_day ) {
               $startAggrTime = $currentRecord->{timestamp};
               # $unixdate_counter++;
               $time_complete = &UnixDate("epoch " . $currentRecord->{timestamp}, "%Y-%m-%d %H:%M:%S");
          }




          $last_record_day = $currentRecord->{rec_term_day};
          if ( $currentRecord->{ipdr} ) {
                  $totalIpdr++;
          }

          foreach my $deviceType ( keys %$deviceMetrics ) {
                if ( $deviceMetrics->{$deviceType}->{filter}->() ) {
                     $subelement = _getSubelementName($deviceType);
                     if ( $subelement !~ /Unknown/ && $subelement ) {
                          my $properties;
                          foreach my $prop ( keys %{$deviceMetrics->{$deviceType}->{properties}} ) {
                              my $check = sprintf($deviceMetrics->{$deviceType}->{properties}->{$prop});
                              if ( $check =~ /CODE/ ) {
                                  # property is defined as a subroutine.
                                  $properties .= $prop . "=" . $deviceMetrics->{$deviceType}->{properties}->{$prop}->() . "&";
                              } else {
                                  $properties .= $prop . "=" . $deviceMetrics->{$deviceType}->{properties}->{$prop} . "&";
                              }
                          }
                          for ( my $d = 0; $d < scalar(@{$deviceMetrics->{$deviceType}->{metricArray}}); $d++ ) {
                              $val = $functionMap->{$deviceMetrics->{$deviceType}->{metricArray}->[$d]}->{method}();
                              if ( $val ) {
                                  $metric = $deviceMetrics->{$deviceType}->{formulaPath} . "~" . $deviceMetrics->{$deviceType}->{metricArray}->[$d];
                                  $aggr = $functionMap->{$deviceMetrics->{$deviceType}->{metricArray}->[$d]}->{aggr};

                                  $last_timestamp = $currentRecord->{timestamp};

                                  if ( ! $time_complete ) {
                                       # $unixdate_counter++;
                                       $time_complete = &UnixDate("epoch " . $currentRecord->{timestamp}, "%Y-%m-%d %H:%M:%S");
                                  }

                                  push(@{$aggrHash->{$time_complete}->{$element}->{$subelement}->{$properties}->{$metric}->{values}}, $val);
                                  $aggrHash->{$time_complete}->{$element}->{$subelement}->{$properties}->{$metric}->{aggr} = $aggr;
                                  $aggrCounter++;

                              }
                          }
                     }
                }
          }
          my $del = delete $recordHash->{$currentRecord->{uci}};
          $currentRecord = undef;
    }

    print "Total IPDR variables reports: $totalIpdr : IPDR records...\n";

    print "   recordHash currently has: " . scalar(keys %$recordHash) . " records.\n";
    print "Performing aggregations... " . time() . "\n";
    print "$aggrCounter records put into the aggregation hash...\n";
    # print "Total times UnixDate was called: $unixdate_counter \n";
    # print "Total times calling UnixDate was skipped: $flush_skip_counter \n";

    foreach my $time_c ( sort keys %$aggrHash ) {
        foreach my $ele ( keys %{$aggrHash->{$time_c}} ) {
             foreach my $se ( keys %{$aggrHash->{$time_c}->{$ele}} ) {
                  foreach my $prop ( keys %{$aggrHash->{$time_c}->{$ele}->{$se}} ) {
                       foreach my $metr ( keys %{$aggrHash->{$time_c}->{$ele}->{$se}->{$prop}} ) {
                            my ( $aggr, $val );
                            ##  do the aggregations
                            my $vals = $aggrHash->{$time_c}->{$ele}->{$se}->{$prop}->{$metr}->{values};
                            my $numEntries = scalar(@{$vals});
                            my $sumval;
                            $aggr = $aggrHash->{$time_c}->{$ele}->{$se}->{$prop}->{$metr}->{aggr};
                            if ( $aggr eq "average" ) {
                                 my $tmpval;
                                 map { $tmpval += $_ } @$vals;
                                 $val = $tmpval / $numEntries;
                            } else {
                                 map { $val += $_ } @$vals;
                            }
                            $datafile->[$filecounter - 1] .= join (',', ( $time_c,$ele,$se,$prop,$metr,$aggr,$val )) . "\n";
                            if ( $filecounter == $numberOfOutputs ) {
                                 $filecounter = 1;
                            } else {
                                 $filecounter++;
                            }
                       }
                  }
             }
        }
        delete $aggrHash->{$time_c};
        $time_c = undef;
    }

    print "Aggregations complete..." . time() . "\n";

    if ( $datafile ) {
       # my $time = &UnixDate("epoch $last_timestamp", "%Y_%m_%d_%H_%M_%S");
       my $time = &UnixDate("epoch $lastAMAstamp", "%Y_%m_%d_%H_%M_%S");
       for ( my $f = 0; $f < scalar(@$datafile); $f++ ) {
          print "Flushing to file: $outputDir/CS2K_$time.$f.csv at " . time() . "\n";
          open(OUTPUTFILE, "> $outputDir/CS2K_$time.$f.csv");
          print OUTPUTFILE $datafile->[$f];
          close(OUTPUTFILE);
       }
    }
    my $ipdrRemove = 0;
    print "Total records in hash before delete: " . scalar(keys %$recordHash) . "\n";
    my $dup_counter = 0;
    my $ama_count = 0;
    my $ipdr_count = 0;
    my $other_count = 0;
    foreach my $rec ( keys %$recordHash ) {
        # these will be all IPDR at this point.  check to see if any should be aged out.
        $recordHash->{$rec}->{poll}++;
        if ( $recordHash->{$rec}->{poll} > 10 && ! $recordHash->{$rec}->{ama} ) {
             delete $recordHash->{$rec};
             $ipdrRemove++;
        } else {
             if ( $DEBUG ) {
                if ( $recordHash->{$rec}->{ama} ) {
                     $ama_count++;
                } elsif ( $recordHash->{$rec}->{ipdr} ) {
                     $ipdr_count++;
                } else {
                     $other_count++;
                }
            }
        }
    }
    print "Done.  Removed $ipdrRemove records.   Removed $dup_counter dup records.\n";


    print ">>> TIMESTAMPEND: " . time() . "\n";
    if ( $DEBUG ) {
       print "Number of total records still in memory after flush: " . scalar(keys %$recordHash) . "\n";
       print "     AMA records: " . scalar(@$ama_array) . "\n";
       print "     IPDR records: $ipdr_count\n";
       print "     OTHER?? records:  $other_count\n";
       # print "Number of AMA records in linked list still in memory after flush: $ama_counter\n";
    }
}







sub _generateCallId( $ ) {
     my $call = shift;
     my $hexCall;
     my $callArr = [ split(//, $call) ];
     for ( my $i = 0; $i < length($call); $i = $i + 3 ) {
          my $hex = sprintf("%x", $callArr->[$i] . $callArr->[$i + 1] . $callArr->[$i + 2]);
          if ( length($hex) == 1 ) {
            $hex = "0" . $hex;
          }
          $hexCall .= $hex;
     }
     return $hexCall;
}




#########################################
##### IPDR handling


sub _processIPDR() {

   $logType = "ipdr";
   my $newFileArr;
   $aboveMOScounter = 0;
   $qosCallCounter = 0;

   opendir(IPDR_STORE, $ipdrdir) || die "Could not open dir '$ipdrdir': $!\n";
   my @old_files = readdir(IPDR_STORE);

   #foreach my $host ( keys %$xmlHosts ) {
   #    print "Getting IPDR files from URI: $host ...\n";
   #   my $ftp = new Net::FTP( $host );
   #  $ftp->login( $xmlHosts->{$host}->{user}, $xmlHosts->{$host}->{pass} );
   # $ftp->cwd( $xmlHosts->{$host}->{path} );
   # my $dir = $ftp->dir();

   opendir(NEW_FILE, $ipdrpath) || die "Could not open dir '$ipdrpath': $!\n";
   my @dir = readdir(NEW_FILE);
   #my $dir = \@dirip ;
   #print "FOO: $dir. \n";
   my $fileArray;
   my $fcounter = 0;
   # map { $_ = (split(/\s+/, $_))[0] } @dir;
   foreach my $listFile ( sort _ipdrsort grep { $_ ne '.' && $_ ne '..' } @dir  ) {
       push(@$fileArray, $listFile) if $listFile;
   }

   foreach my $f ( @$fileArray ) {
       if ( ! grep { $_ eq $f } @old_files ) {
            if ( $fcounter < $maxFilesPerRun ) {
               # my $window = 15000;  ideally, this should be 900 or 1800 seconds
               my $window = 7200; # testing
               if ( _ipdrstamp($f) >= ( $lastAMAstamp - $window ) &&  _ipdrstamp($f) <= ( $lastAMAstamp + $window ) ) {
                  # we have a new file...
                  #  $ftp->binary();
                  print "Downloading file: $ipdrpath/$f ... ";
                  move("$ipdrpath/$f", "$ipdrdir/$f");
                  print "Done.\n";
                  push ( @$newFileArr, $f );
                  $fcounter++;
               }
            }
       } else {
            print "Skipping already obtained file: $f\n";
       }

   }


   my $timestampStart = time();
   if ( $newFileArr ) {
      print "Processing IPDR files.... TIMESTAMP: $timestampStart\n";
   }



   my $ipdr_date_counter = 0;
   my $ipdr_skip_counter = 0;
   foreach my $file ( @$newFileArr ) {
      print "Processing file: $file at " . time() . "\n";
      my ( $data, $foo );
      open (IPDRFILE, "gzip -dc $ipdrdir/$file |") || die "Unable to open file '$file': $!\n";
      my $lineCount = 0;
      my $record;
      foreach my $line (<IPDRFILE>) {
         $lineCount++;
         if ( $line =~ /IPDR>/ ) {
             # new record
             if ( $record ) {
                # we have a complete record. flush it to the database.
                # .. logic logic
                # print STDERR Dumper($record) . "\n";
                if ( $record->{MOSLQ} ) {
                    $qosCallCounter++;
                    if ( $record->{MOSLQ} > $MOSsla ) {
                        $aboveMOScounter++;
                    }
                }
                foreach my $key ( keys %$record ) {
                   $recordHash->{$record->{uniqueCallId}}->{$key} = $record->{$key};
                }
                $recordHash->{$record->{uniqueCallId}}->{ipdr} = 1;
                $recordHash->{$record->{uniqueCallId}}->{uci} = $record->{uniqueCallId};
                $record = {};
             }
         } elsif ( $line !~ /IPDRDoc/ && $line !~ /xml/ ) {
             my ( $tag, $value  ) = $line =~ /<(.*)>(.*)</;
             if ( $tag ) {
                 $record->{$tag} = $value;
             }
         }
      }
      print "Completed processing file: $file at " . time() . "\n";
      print " Called Date::Manip $ipdr_date_counter times and skipped it $ipdr_skip_counter times\n";
   }

}



################################################################################
################################################################################
################################################################################
##### DO NOT EDIT BELOW THIS LINE ##############################################
################################################################################
################################################################################
################################################################################



my $pid = fork();


if ( $pid ) {
  # Watchdog routine
  #
  while ( 1 ) {
     sleep 30;
     my $data = `/bin/ps -efao pmem -o fname -o rss -o vsz | grep cs2k`;
     my $pid = `/bin/ps -eaf | grep perl | grep cs2k | awk '{print \$2}'`;
     my $parr = [ split(/\s+/, $pid) ];
     if ( scalar(@$parr) < 2 ) {
        # main isn't running...
        system("$launchpath");
        exit;
     }
  }
}


setsid;

$| = 1;


while ( 1 ) {
   $startEpoch = time();
   &_processIPDR();
   my $p = &_processAMA();
   if ( $p ) {
      &_flushData();
   }
   my $timeNow = time();
   my $sleep = ( $startEpoch + $runInterval ) - $timeNow;
   if ( $sleep > 0 ) {
      print "Sleeping for $sleep\n";
      sleep $sleep;
   }
}

